[synopsis]
crm_mon mode [options]

/number of different formats/
.SH OPTIONS

/less descriptive in output./
.SH NOTES

/period specification./
.SH OUTPUT CONTROL

/command line arguments./
.SH TIME SPECIFICATION

/or --exclude=list./
.SH INTERACTIVE USE
